from storm.variables import Variable

from spans import *

__all__ = [
    "RangeVariable",
    "IntRangeVariable",
    "FloatRangeVariable",
    "DateRangeVariable",
    "DateTimeRangeVariable"
]

class RangeVariable(Variable):
    """
    Extension of standard variable class to handle conversion to and from Psycopg2
    range types
    """

    def parse_set(self, value, from_db):
        if not isinstance(value, self.range_type):
            raise ValueError(
                "Expected '{range_type.__name__}' '{value!r}' given".format(
                    range_type=self.range_type,
                    value=value))

        return value

    def parse_get(self, value, to_db):
        return value

# Define the builtin range properties and variables
class IntRangeVariable(RangeVariable):
    range_type = intrange

class FloatRangeVariable(RangeVariable):
    range_type = floatrange

class DateRangeVariable(RangeVariable):
    range_type = daterange

class DateTimeRangeVariable(RangeVariable):
    range_type = datetimerange



from datetime import date, datetime, timedelta
from os import environ
from psycopg2 import ProgrammingError
from spans import *
from storm.expr import State
from storm.locals import *
from storm.variables import ListVariable
from unittest import TestCase, skipUnless

from .. import *
from ..database import install_range
from ..expr import *
from ..properties import RangeProperty
from ..variables import \
    RangeVariable, \
    IntRangeVariable, \
    FloatRangeVariable, \
    DateRangeVariable, \
    DateTimeRangeVariable

@skipUnless(
    environ.get("STORM_POSTGRES_URI"),
    "No Storm Postgres URI provided, skipping database tests")
class TestDatabase(TestCase):
    def setUp(self):
        self.store = Store(create_database(environ["STORM_POSTGRES_URI"]))
        self.create_tables()
        self.fill_data()
        self.store.commit()

    def tearDown(self):
        self.store.rollback()
        self.drop_tables()
        self.store.commit()

    def create_tables(self):
        pass

    def fill_data(self):
        pass

    def drop_tables(self):
        pass

class TestOperators(TestDatabase):
    def run_operator(self, operator, arg1, arg2):
        return self.store.execute(Select(operator(arg1, arg2))).get_one()[0]

    def test_contains(self):
        self.assertTrue(self.run_operator(
            Contains, intrange(1, 5), intrange(1, 5)))
        self.assertFalse(self.run_operator(
            Contains, intrange(1, 5), intrange(1, 5, upper_inc=True)))

    def test_within(self):
        self.assertTrue(self.run_operator(
            Within, intrange(1, 5), intrange(1, 5, upper_inc=True)))
        self.assertFalse(self.run_operator(
            Within, intrange(1, 5, upper_inc=True), intrange(1, 5)))

    def test_overlap(self):
        self.assertTrue(self.run_operator(
            Overlap, intrange(1, 5), intrange(3, 8)))
        self.assertFalse(self.run_operator(
            Overlap, intrange(1, 5), intrange(5, 10)))

    def test_adjacent(self):
        self.assertFalse(self.run_operator(
            Adjacent, intrange(1, 5), intrange(3, 8)))
        self.assertTrue(self.run_operator(
            Adjacent, intrange(1, 5), intrange(5, 10)))

    def test_startsafter(self):
        self.assertFalse(self.run_operator(
            StartsAfter, intrange(1, 5), intrange(3, 8)))
        self.assertTrue(self.run_operator(
            StartsAfter, intrange(1, 5), intrange(1, 5)))
        self.assertTrue(self.run_operator(
            StartsAfter, intrange(5, 10), intrange(1, 5)))

    def test_endsbefore(self):
        self.assertTrue(self.run_operator(
            EndsBefore, intrange(1, 5), intrange(3, 8)))
        self.assertTrue(self.run_operator(
            EndsBefore, intrange(1, 5), intrange(1, 5)))
        self.assertFalse(self.run_operator(
            EndsBefore, intrange(5, 10), intrange(1, 5)))

class TestFunctions(TestDatabase):
    def run_expr(self, expr):
        return self.store.execute(Select(expr)).get_one()[0]

    def test_lower_inc(self):
        self.assertTrue(self.run_expr(LowerInc(intrange(1, 5))))
        self.assertTrue(self.run_expr(LowerInc(intrange(1, 5, lower_inc=False))))
        self.assertFalse(self.run_expr(LowerInc(floatrange(1.0, 5.0, lower_inc=False))))

    def test_upper_inc(self):
        self.assertFalse(self.run_expr(UpperInc(intrange(1, 5))))
        self.assertFalse(self.run_expr(UpperInc(intrange(1, 5, upper_inc=True))))
        self.assertTrue(self.run_expr(UpperInc(floatrange(1.0, 5.0, upper_inc=True))))

    def test_lower_inf(self):
        self.assertTrue(self.run_expr(LowerInf(intrange(upper=5))))
        self.assertFalse(self.run_expr(LowerInf(intrange(1))))

    def test_upper_inf(self):
        self.assertFalse(self.run_expr(UpperInf(intrange(upper=5))))
        self.assertTrue(self.run_expr(UpperInf(intrange(1))))

class TestIntRangeDatabase(TestDatabase):
    class SimpleModel(Storm):
        __storm_table__ = "int_range_test"

        id = Int(primary=True, default=AutoReload)
        span = IntRange(default=intrange(10, 15))

        def __init__(self, span=None):
            if span is not None:
                self.span = span

    def create_tables(self):
        self.store.execute("""
            CREATE TABLE int_range_test(
                id SERIAL,
                span int4range
            )
        """)

    def fill_data(self):
        self.store.execute("""
            INSERT INTO int_range_test(span) VALUES (?),(?),(?)
        """, [intrange(1, 5), intrange(3, 8), intrange(5, 10)])

    def drop_tables(self):
        self.store.execute("""
            DROP TABLE IF EXISTS
                int_range_test
                CASCADE
        """)

    def test_get(self):
        ir = self.store.get(self.SimpleModel, 1)

        self.assertIsNot(ir, None)
        self.assertEquals(ir.span, intrange(1, 5))

    def test_default_insert(self):
        ir = self.SimpleModel()
        self.store.add(ir)
        self.store.commit()
        self.store.invalidate(ir)

        self.assertEquals(ir.span, intrange(10, 15))

    def test_insert(self):
        ir = self.SimpleModel(intrange(upper=10, lower_inc=False))
        self.store.add(ir)
        self.store.commit()
        self.store.invalidate(ir)

        self.assertEquals(ir.span, intrange(upper=10, lower_inc=False))

    def test_simple_functions(self):
        self.assertEqual(
            list(self.store.find(
                self.SimpleModel.span.lower()).order_by(self.SimpleModel.id)),
            [1, 3, 5])
        self.assertEqual(
            list(self.store.find(
                self.SimpleModel.span.upper()).order_by(self.SimpleModel.id)),
            [5, 8, 10])

        self.assertEqual(
            list(self.store.find(
                self.SimpleModel.span.lower_inc()).order_by(self.SimpleModel.id)),
            [True] * 3)
        self.assertEqual(
            list(self.store.find(
                self.SimpleModel.span.upper_inc()).order_by(self.SimpleModel.id)),
            [False] * 3)

        self.assertEqual(
            list(self.store.find(
                self.SimpleModel.span.lower_inf()).order_by(self.SimpleModel.id)),
            [False] * 3)
        self.assertEqual(
            list(self.store.find(
                self.SimpleModel.span.upper_inf()).order_by(self.SimpleModel.id)),
            [False] * 3)

    def test_contains(self):
        self.assertEqual(self.store.find(
            self.SimpleModel,
            self.SimpleModel.span.contains(1)).count(), 1)
        self.assertEqual(self.store.find(
            self.SimpleModel,
            self.SimpleModel.span.contains(3)).count(), 2)
        self.assertEqual(self.store.find(
            self.SimpleModel,
            self.SimpleModel.span.contains(5)).count(), 2)
        self.assertEqual(self.store.find(
            self.SimpleModel,
            self.SimpleModel.span.contains(10)).count(), 0)

        self.assertEqual(self.store.find(
            self.SimpleModel,
            self.SimpleModel.span.contains(intrange(1, 3))).count(), 1)
        self.assertEqual(self.store.find(
            self.SimpleModel,
            self.SimpleModel.span.contains(intrange(3, 5))).count(), 2)
        self.assertEqual(self.store.find(
            self.SimpleModel,
            self.SimpleModel.span.contains(intrange(5, 8))).count(), 2)
        self.assertEqual(self.store.find(
            self.SimpleModel,
            self.SimpleModel.span.contains(intrange(10, 15))).count(), 0)

class TimeDeltaRangeVariable(RangeVariable):
    range_type = timedeltarange

class TimeDeltaRange(RangeProperty):
    variable_class = TimeDeltaRangeVariable

class CustomTypeModel(Storm):
    __storm_table__ = "td_range_test"

    id = Int(primary=True)
    span = TimeDeltaRange()
    spans = List(type=TimeDeltaRange())

class TestCustomRangeDatabase(TestDatabase):
    class SimpleModel(Storm):
        __storm_table__ = "int_range_test"

        id = Int(primary=True, default=AutoReload)
        span = IntRange(default=intrange(10, 15))

        def __init__(self, span=None):
            if span is not None:
                self.span = span

    def setUp(self):
        super(TestCustomRangeDatabase, self).setUp()

        # Create tables
        try:
            self.store.execute("""
                CREATE TYPE intervalrange AS RANGE(SUBTYPE = interval)
            """)
        except ProgrammingError:
            # The type already exists in this database
            self.store.rollback()

        self.store.execute("""
            CREATE TABLE IF NOT EXISTS td_range_test(
                id SERIAL,
                span intervalrange,
                spans intervalrange[]
            )
        """)

        self.store.commit()

        install_range("intervalrange", TimeDeltaRangeVariable, self.store)

        self.store.execute(
            "INSERT INTO td_range_test(span, spans) VALUES (?, ?)",
            [
                timedeltarange(timedelta(days=1), timedelta(days=5)),
                [timedeltarange(timedelta(days=1), timedelta(days=2))]])
        self.store.commit()

    def drop_tables(self):
        self.store.execute("""
            DROP TABLE IF EXISTS
                td_range_test
                CASCADE
        """)
        self.store.execute("""DROP TYPE intervalrange CASCADE""")

    def test_get(self):
        r = self.store.get(CustomTypeModel, 1)

        self.assertEqual(
            r.span, timedeltarange(timedelta(days=1), timedelta(days=5)))
        self.assertEqual(
            r.spans, [timedeltarange(timedelta(days=1), timedelta(days=2))])

    def test_insert(self):
        r = CustomTypeModel()
        r.span = timedeltarange(timedelta(days=1), timedelta(days=10))
        r.spans = [
            timedeltarange(timedelta(days=1), timedelta(days=5)),
            timedeltarange(timedelta(days=5), timedelta(days=10))]
        self.store.add(r)
        self.store.commit()

        self.store.invalidate(r)
        self.assertEqual(
            r.span, timedeltarange(timedelta(days=1), timedelta(days=10)))
        self.assertEqual(
            r.spans,
            [
                timedeltarange(timedelta(days=1), timedelta(days=5)),
                timedeltarange(timedelta(days=5), timedelta(days=10))])




"""
StormSpans brings support for PostgreSQL's range types to Canonical's Storm
using PsycoSpans paired with Spans.

    from spans import intrange
    from storm.locals import *
    from stormspans import IntRange

    class Model(Storm):
        id = Int(primary=True)
        span = IntRange(default=intrange(1, 10))

To connect to the database "postgres+spans://..." must be specified instead of
"postgres://..."
"""

__version__ = "0.1.0"

from storm.database import register_scheme

from .database import PostgresStormSpans, install_range
from .properties import *

__all__ = [
    "IntRange",
    "FloatRange",
    "DateRange",
    "DateTimeRange",
    "install_range"
]

register_scheme("postgres+spans", PostgresStormSpans)



from storm.expr import NamedFunc, BinaryOper

__all__ = [
    "Within",
    "Contains",
    "Overlap",
    "Adjacent",
    "StartsAfter",
    "EndsBefore",

    "LowerInc",
    "UpperInc",
    "LowerInf",
    "UpperInf"
]

class Within(BinaryOper):
    __slots__ = ()
    oper = " <@ "

class Contains(BinaryOper):
    __slots__ = ()
    oper = " @> "

class Overlap(BinaryOper):
    __slots__ = ()
    oper = " && "

class Adjacent(BinaryOper):
    __slots__ = ()
    oper = " -|- "

class StartsAfter(BinaryOper):
    __slots__ = ()
    oper = " &> "

class EndsBefore(BinaryOper):
    __slots__ = ()
    oper = " &< "

class LowerInc(NamedFunc):
    __slots__ = ()
    name = "lower_inc"

class UpperInc(NamedFunc):
    __slots__ = ()
    name = "upper_inc"

class LowerInf(NamedFunc):
    __slots__ = ()
    name = "lower_inf"

class UpperInf(NamedFunc):
    __slots__ = ()
    name = "upper_inf"



from storm.expr import Column, Undef, Add, Sub, Mul, LShift, RShift, Lower, Upper
from storm.locals import *
from storm.properties import Property, PropertyColumn

from .variables import *
from .expr import *

__all__ = [
    "IntRange",
    "FloatRange",
    "DateRange",
    "DateTimeRange"
]


class RangeProperty(Property):
    def __init__(self, name=None, primary=False, type=None, **kwargs):
        if type is None:
            type = Property()

        kwargs["value"] = kwargs.pop("default", Undef)
        kwargs["value_factory"] = kwargs.pop("default_factory", Undef)

        Property.__init__(self, name, primary, self.variable_class, kwargs)

    def _get_column(self, cls):
        # Cache per-class column values in the class itself, to avoid
        # holding a strong reference to it here, and thus rendering
        # classes uncollectable in certain situations (e.g. subclasses
        # where the property is stored in the base).
        try:
            # Use class dictionary explicitly to get sensible
            # results on subclasses.
            column = cls.__dict__["_storm_columns"].get(self)
        except KeyError:
            cls._storm_columns = {}
            column = None
        if column is None:
            attr = self._detect_attr_name(cls)
            if self._name is None:
                name = attr
            else:
                name = self._name
            column = RangePropertyColumn(self, cls, attr, name, self._primary,
                                    self._variable_class,
                                    self._variable_kwargs)
            cls._storm_columns[self] = column
        return column

class RangeComparable(object):
    def lower(self):
        return Lower(self)

    def upper(self):
        return Upper(self)

    def lower_inc(self):
        return LowerInc(self)

    def upper_inc(self):
        return UpperInc(self)

    def lower_inf(self):
        return LowerInf(self)

    def upper_inf(self):
        return UpperInf(self)

    def contains(self, other):
        return Contains(self, other)

    def within(self, other):
        return Within(self, other)

    def overlap(self, other):
        return Overlap(self, other)

    def adjacent(self, other):
        return Adjacent(self, other)

    def union(self):
        return Add(self, other)

    def difference(self, other):
        return Sub(self, other)

    def intersection(self, other):
        return Mul(self, other)

    def left_of(self, other):
        return LShift(self, other)

    def right_of(self, other):
        return RShift(self, other)

    def __contains__(self, other):
        return Contains(self, other)

    def __lshift__(self, other):
        return self.left_of(other)

    def __rshift__(self, other):
        return self.right_of(other)

class RangePropertyColumn(RangeComparable, PropertyColumn):
    pass

class IntRange(RangeProperty):
    variable_class = IntRangeVariable

class FloatRange(RangeProperty):
    variable_class = FloatRangeVariable

class DateRange(RangeProperty):
    variable_class = DateRangeVariable

class DateTimeRange(RangeProperty):
    variable_class = DateTimeRangeVariable



import os
import imp
import unittest

from glob import iglob

def find_tests():
    for path in iglob(os.path.dirname(os.path.realpath(__file__)) + "/test_*.py"):
        name = os.path.basename(path)[:-3]
        yield unittest.defaultTestLoader.loadTestsFromModule(
            imp.load_source(__name__ + "." + name, path))

def test_suite():
    return unittest.TestSuite(find_tests())

def main():
    runner = unittest.TextTestRunner(verbosity=1)
    return runner.run(unittest.TestSuite(find_tests()))



#!/usr/bin/env python
import re
import sys
import os

from setuptools import setup

with open("README.rst") as fp:
    long_desc = fp.read()

with open("LICENSE") as fp:
    license = fp.read()

with open("stormspans/__init__.py") as fp:
    version = re.search('__version__\s+=\s+"([^"]+)', fp.read()).group(1)

requirements = []
if os.path.exists("requirements.txt"):
    with open("requirements.txt") as fp:
        requirements = fp.read().split("\n")

setup(
    name="StormSpans",
    version=version,
    description=
        "PostgreSQL range type support for Canonical's Storm ORM using Spans ",
    long_description=long_desc,
    license=license,
    author="Andreas Runfalk",
    author_email="andreas@runfalk.se",
    url="https://www.github.com/runfalk/stormspans",
    packages=["stormspans"],
    install_requires=requirements,
    classifiers=(
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Topic :: Database",
        "Topic :: Database :: Front-Ends"
    ),
    zip_safe=False,
    test_suite="stormspans.tests.test_suite"
)



from storm.databases.postgres import Postgres, PostgresConnection, compile as postgres_compile
from psycospans import connect, register_range_type
from spans import *
from spans.types import range_ as spansbase

__all__ = [
    "install_range"
]

compile = postgres_compile.create_child()

@compile.when(spansbase)
def compile_range(compile, range, state):
    state.parameters.append(range)
    return "?"

class PostgresStormSpansConnection(PostgresConnection):
    compile = compile

class PostgresStormSpans(Postgres):
    connection_factory = PostgresStormSpansConnection

    def raw_connect(self):
        raw_connection = connect(self._dsn)

        if self._version is None:
            self._version = raw_connection.server_version

        raw_connection.set_client_encoding("UTF8")
        raw_connection.set_isolation_level(self._isolation)
        return raw_connection

def install_range(pgrange, pyrange_variable, store):
    register_range_type(
        pgrange,
        pyrange_variable.range_type,
        store._connection._raw_connection)
